import pandas as pd
from requests import Response

from .explain_utils import get_scores_summary, get_scores_terms
from .types import (
    CatDict,
    CountDict,
    ExplainDetailsDict,
    FieldScoreDict,
    ScoreSummaryDict,
    StatsResultsDict,
)


class ExplainResult:
    """ExplainResult class

    Methods
    -------
    These methods are available:
        - `self.get_scores_breakdown(as_df: bool = False)`
        - `self.get_scores_term(as_df: bool = False)`
        - `self.get_scores_summary(as_df: bool = False)`

    Attributes
    ----------
    These attributes are available:
        - `status_code`
        - `json`
        - `explanation`
        - `score`
    """

    def __init__(self, data: Response | ExplainDetailsDict):
        if isinstance(data, Response):
            self._status_code = data.status_code
            self._response = data
            self._json = data.json()

        elif isinstance(data, dict):  # explanation dict from search results
            if all(key in data for key in ("value", "description", "details")):
                self._status_code = 200
                self._response = None
                self._json = {"explanation": data}

        else:
            raise TypeError("Unsupported input type for ExplainResult")

    @property
    def status_code(self) -> int:
        """Returns the `status_code`"""
        return self._status_code

    @property
    def json(self) -> dict:
        """Returns the json data"""
        return self._json

    @property
    def explanation(self) -> dict:
        """Returns the explanation"""
        return self.json.get("explanation", {})

    @property
    def score(self) -> float | None:
        """Returns the final score"""
        return self.explanation.get("value")

    def get_scores_breakdown(self, as_df: bool = False) -> list | pd.DataFrame:
        """Gets a breakdown of the scores structure with depth and its information

        Parameters
        ----------
            - as_df (bool): Whether or not to return the results in DataFrame (Default = False)

        Returns
        -------
            list | pd.DataFrame
        """

        result = []

        def walk(node, depth=0):
            desc = node.get("description", "")
            val = node.get("value", 0)
            result.append((depth, val, desc))
            for detail in node.get("details", []):
                walk(detail, depth + 1)

        walk(self.explanation)
        if as_df:
            return pd.DataFrame(result, columns=["depth", "score", "description"])

        return result

    def get_scores_term(
        self, as_df: bool = False
    ) -> list[FieldScoreDict] | pd.DataFrame:
        """Returns the tf, idf, boost for each field

        Parameters
        ----------
            - as_df (bool): Whether or not to return the results in DataFrame (Default = False)

        Returns
        -------
            list[FieldScoreDict] | pd.DataFrame
        """

        list_score_terms = get_scores_terms(self.explanation)

        if as_df:
            return pd.DataFrame(list_score_terms)

        return list_score_terms

    def get_scores_summary(
        self, as_df: bool = False
    ) -> dict[str, ScoreSummaryDict] | pd.DataFrame:
        """Returns the summary for each term's contributions

        Parameters
        ----------
            - as_df (bool): Whether or not to return the summary in DataFrame (Default = False)

        Returns
        -------
            dict[str, ScoreSummaryDict] | pd.DataFrame
        """

        list_score_terms = self.get_scores_term()
        score_summary_results = get_scores_summary(list_score_terms)

        if as_df:
            df = pd.DataFrame(score_summary_results).transpose()
            df.index.name = "field"

            return df

        return score_summary_results

    def __repr__(self):
        return f"<ExplainResult explanation_value={self.score}>"


class SearchResults:
    """SearchResults class

    Methods
    -------
    These methods are available:
        - `self.get_sources(include_id: bool = False, include_score: bool = False, as_list: bool = False)`
        - `self.get_ids()`
        - `self.get_explanations()`
        - `self.to_dataframe(columns: list[str], include_id: bool = False, include_score: bool = False)`

    Attributes
    ----------
    These attributes are available:
        - `status_code`
        - `json`
        - `hits`
        - `total`
    """

    def __init__(self, response: Response):
        self._status_code = response.status_code
        self._json: dict = response.json()

    @property
    def status_code(self) -> int:
        """Returns status code from requests"""
        return self._status_code

    @property
    def json(self) -> dict:
        """Returns the json data"""
        return self._json

    @property
    def hits(self) -> list[dict | None]:
        """Returns the `hits` dict of the results"""
        return self.json.get("hits", {}).get("hits", [])

    @property
    def total(self) -> int:
        """Returns a total count"""

        return self.json.get("hits", {}).get("total", {}).get("value", 0)

    def get_total_hits(self):
        """Returns a total count"""
        return self.total

    def get_hits(self) -> list[dict | None]:
        """Returns the `hits` dict of the results"""
        return self.hits

    def get_sources(
        self,
        include_id: bool = False,
        include_score: bool = False,
        as_list: bool = False,
    ) -> dict[str, dict] | list[dict | None]:
        """Returns _source dict results

        Parameters
        ----------
            - include_id (bool): Whether or not to include `_id` in the results

            - include_score (bool): Whether or not to include `_score` in the results

            - as_list (bool): Whether or not to return as a list or dict, default is dict (Default = False)

        Returns
        -------
            dict[str, dict] | list[dict | None]
        """

        if not include_id and not include_score:
            sources_dict = {hit["_id"]: hit["_source"] for hit in self.hits}
            return sources_dict if not as_list else list(sources_dict.values())

        sources_dict = {}
        for hit in self.hits:
            item = hit["_source"].copy()
            if include_id:
                item["_id"] = hit["_id"]
            if include_score:
                item["_score"] = hit["_score"]

            sources_dict[hit["_id"]] = item

        return sources_dict if not as_list else list(sources_dict.values())

    def get_ids(self) -> list[str]:
        """Returns a list of documents' `_id`"""

        return [hit["_id"] for hit in self.hits]

    def get_explanations(self) -> dict[str, ExplainResult] | None:
        """Returns _explanation results

        If `_explanation` exists in the results,
        returns a dictionry of ExplainResult with
        the key being the `_id`

        Example
        -------
            The results will be in a form of:
            ```
            {"xx": <ExplainResult explanation_value=...>,
             "yy": <ExplainResult explanation_value=...>,}
            ```
        """

        if self.hits[0].get("_explanation"):
            return {hit["_id"]: ExplainResult(hit["_explanation"]) for hit in self.hits}

    def to_dataframe(
        self,
        columns: list[str] | None = None,
        include_id: bool = False,
        include_score: bool = False,
    ) -> pd.DataFrame:
        """Returns pandas DataFrame object

        Parameters
        ----------
            - columns (list[str] | None): Columns in the DataFrame results, if None, returns every field (Default = None)

            - include_id (bool): Whether or not to include `_id` in the results

            - include_score (bool): Whether or not to include `_score` in the results

        Returns
        -------
            pd.DataFrame
        """

        df = pd.DataFrame(
            self.get_sources(
                include_id=include_id, include_score=include_score, as_list=False
            )
        ).transpose()

        df.index.name = "_id"

        if columns:
            return df[columns]

        return df

    def __repr__(self):
        return f"<SearchResults total_hits={self.total}>"


class CountResults:
    def __init__(self, response: Response):
        self._status_code = response.status_code
        self._json: CountDict = response.json()

    @property
    def status_code(self) -> int:
        """Returns status code from requests"""
        return self._status_code

    @property
    def json(self) -> CountDict:
        """Returns the json data"""
        return self._json

    @property
    def count(self) -> int:
        """Returns the documents count"""
        return self.json["count"]

    def __repr__(self):
        return f"<CountResults count={self.count}>"


class CatResults:
    def __init__(self, response: Response):
        self._status_code = response.status_code
        self._json: list[dict] = response.json()
        self._json = [self._normalise_cat_keys(row) for row in self._json]

    @property
    def status_code(self) -> int:
        """Returns status code from requests"""
        return self._status_code

    @property
    def json(self) -> list[CatDict]:
        """Returns the json data"""
        return self._json

    @property
    def total(self) -> int:
        """Returns the count"""
        return len(self.json)

    def get_total_indices(self):
        """Returns the total indices count"""
        return self.total

    def get_indices(self) -> list[str]:
        """Returns a list of all the indices"""
        return [row["index"] for row in self.json]

    def filter_indices(self, indices: str | list[str]):
        """Filters the results to the specified index name(s) only"""

        def _filter_results(row):
            if isinstance(indices, str):
                return row["index"] == indices
            elif isinstance(indices, list):
                return row["index"] in indices
            else:
                raise ValueError("indices must be either str or list[str]")

        return list(filter(_filter_results, self.json))

    @staticmethod
    def _normalise_cat_keys(entry: dict) -> CatDict:
        return {k.replace(".", "_"): v for k, v in entry.items()}

    def __repr__(self):
        return f"<CatResults total_indices={self.total}>"


class StatsResults:
    def __init__(self, response: Response):
        self._status_code = response.status_code
        self._json: dict = response.json()

    @property
    def status_code(self) -> int:
        """Returns status code from requests"""
        return self._status_code

    @property
    def json(self) -> StatsResultsDict:
        """Returns the json data"""
        return self._json

    @property
    def total(self):
        """Returns the total documents count"""
        return self.json["_all"]["total"]["docs"]["count"]

    @property
    def size(self):
        """Returns the total size in bytes for the index"""
        return self.json["_all"]["total"]["store"]["size_in_bytes"]

    def get_total_count(self):
        """Returns the total documents count"""
        return self.total

    def get_total_size(self):
        """Returns the total size in bytes for the index"""
        return self.size

    def get_indices(self) -> list[str]:
        """Get a list of indices within this index (maybe aliases)"""
        return list(self.json["indices"].keys())

    def __repr__(self):
        return f"<StatsResults total_count={self.total}, total_size={self.size}>"



from requests.models import Response


class RequestResponse(Response):
    def __init__(self, status_code: int, content: str):
        super().__init__()
        self.status_code = status_code
        self._content = content.encode("utf-8")  # must be bytes



import base64
import json
import posixpath as ppath
from typing import Literal

import requests

from .response import RequestResponse
from .results import (
    CatResults,
    CountResults,
    SearchResults,
    ExplainResult,
    StatsResults,
)


class ESClient:
    def __init__(
        self,
        es_endpoint: str | None = None,
        api_key: str | None = None,
        basic_auth: tuple[str, str] | None = None,
    ):
        """Initialise Elasticsearch tool object

        Parameters
        ----------
            - es_endpoint (str): URL endpoint of the Elasticsearch cluster

            - api_key (str): API key for authentication (for API key authentication)

            - basic_auth (tuple[str, str]): A tuple of (username, password) (for basic authentication)
        """
        self.es_endpoint = es_endpoint
        self.api_key = api_key
        self.basic_auth = basic_auth
        self.headers = None
        self.is_authen = False

        if self.es_endpoint is not None and self.api_key is not None:
            self.authen(self.es_endpoint, api_key=self.api_key)

        elif self.es_endpoint is not None and self.basic_auth is not None:
            self.authen(self.es_endpoint, basic_auth=self.basic_auth)

        print(self.__repr__())
        if not self.is_authen:
            print(
                "Please authenticate using `authen()` method, or pass in the credentials at initialisation."
            )

    def authen(
        self,
        es_endpoint: str,
        api_key: str | None = None,
        basic_auth: tuple[str, str] | None = None,
    ):
        """Authenticates to the Elasticsearch endpoint

        Supports 2 types of authentication:
            - using an API key (`api_key`)
            - using a username and password (`basic_auth`)

        Parameters
        ----------
            - es_endpoint (str): Elasticsearch endpoint

            - api_key (str): API key for authentication (for API key authentication)

            - basic_auth (tuple[str, str]): A tuple of (username, password) (for basic authentication)

        Returns
        -------
            Response
        """

        if es_endpoint is None:
            raise ValueError("`es_endpoint` must have a value.")

        if api_key is None and basic_auth is None:
            raise ValueError("Either `api_key` or `basic_auth` must be passed in.")

        self.es_endpoint = es_endpoint
        self.api_key = api_key
        self.basic_auth = basic_auth

        if self.api_key:
            self.headers = {"Authorization": f"ApiKey {self.api_key}"}

        elif self.basic_auth:
            if not isinstance(self.basic_auth, tuple) or len(self.basic_auth) != 2:
                raise ValueError("`basic_auth` must be a tuple with a length of 2.")

            user_pass = ":".join(self.basic_auth)
            self.user_pass = base64.b64encode(user_pass.encode("utf-8")).decode()
            self.headers = {"Authorization": f"Basic {self.user_pass}"}

        auth_resp = requests.head(self.es_endpoint, headers=self.headers)

        if auth_resp.status_code == 200:
            self.is_authen = True
            return auth_resp
        else:
            raise Exception(
                f"Authentication failed: {auth_resp.status_code} - {auth_resp.text}"
            )

    def delete_index(self, index_name: str):
        """Deletes an index, if available

        Parameters
        ----------
            - index_name (str): Index name to delete

        Returns
        -------
        Response
        """

        self._check_authen()

        es_url = ppath.join(self.es_endpoint, index_name)

        delete_response = requests.delete(es_url, headers=self.headers)
        if delete_response.status_code == 200:
            print(f"Index '{index_name}' index successfully deleted.")
            return delete_response
        else:
            raise Exception(
                f"Failed to delete index '{index_name}': {delete_response.status_code} - {delete_response.text}"
            )

    def create_index(
        self, index_name: str, json_mapping: dict, replace_if_exists: bool = False
    ):
        """Creates an index with a specified mapping

        Parameters
        ----------
            - index_name (str): Index name to create

            - json_mapping (dict): A dictionary mapping for the index

            - replace_if_exists (bool): Whether or not to delete and re-create a new index if already exists (Default = False)

        Returns
        -------
            Response
        """

        self._check_authen()

        create_headers = self.headers.copy()
        create_headers["Content-Type"] = "application/json"
        es_url = ppath.join(self.es_endpoint, index_name)

        if replace_if_exists:
            # Test whether it already exists
            test_response = requests.head(es_url, headers=self.headers)

            if test_response.status_code == 200:  # exists
                print(f"Found the existing index '{index_name}'")
                # Delete the existing index and create a new one
                _ = self.delete_index(index_name)

        # Create with the mapping
        create_response = requests.put(
            es_url, headers=create_headers, json=json_mapping
        )

        if create_response.status_code == 200:
            print(f"Index '{index_name}' has been created successfully.")

        return create_response

    def index_doc(self, index_name: str, doc_data: dict, _id: str | None = None):
        """Index one document to the specified index

        Parameters
        ----------
        index_name: str
            Index name to put/index data into

        doc_data: dict
            The data in JSON/dict format with the correct fields as the mapping

        _id: str | None (Default = None)
            The _id for the Elasticsearch index

            If None is provided, the `_id` of the document will be auto-generated.

        Returns
        -------
        Response
            The response of this index operation call
        """

        self._check_authen()

        index_headers = self.headers.copy()
        index_headers["Content-Type"] = "application/json"

        es_url = ppath.join(self.es_endpoint, index_name, "_doc")
        if _id:
            es_url += f"/{_id}"
            index_response = requests.put(es_url, headers=index_headers, json=doc_data)

        index_response = requests.post(es_url, headers=index_headers, json=doc_data)

        if index_response.status_code in (200, 201):
            print(f"The data has been successfully indexed into {index_name}")

        return index_response

    def update_doc(
        self,
        index_name: str,
        doc_data: dict,
        _id: str,
        detect_noop: bool = True,
        doc_as_upsert: bool = True,
    ):
        """Update one document to the specified index

        Parameters
        ----------
        index_name: str
            Index name to update data into

        doc_data: dict
            The data in JSON/dict format with the correct fields as the mapping

        _id: str
            The _id to update

            If None is provided, the `_id` of the document will be auto-generated.

        detect_noop: bool (Default = True)
            If `True`, the result in the response is set to noop (no operation) when there are no changes to the document

        doc_as_upsert: bool (Default = True)
            If True, use the contents of `doc_data` as the value of upsert operation

        Returns
        -------
        Response
            The response of this update operation call
        """

        self._check_authen()

        index_headers = self.headers.copy()
        index_headers["Content-Type"] = "application/json"

        es_url = ppath.join(self.es_endpoint, index_name, "_update", _id)

        json_body = {
            "detect_noop": detect_noop,
            "doc_as_upsert": doc_as_upsert,
            "doc": doc_data,
        }

        update_response = requests.post(es_url, headers=index_headers, json=json_body)

        if update_response.status_code == 200:
            print(f"The doc_id: {_id} has been updated at index: {index_name}")

        return update_response

    @staticmethod
    def generate_bulk_payload(
        index_name: str,
        data: list[dict],
        method: Literal["index", "update"] = "index",
        id_key: str | None = None,
    ):
        """Generates bulk payload data, for _bulk API"""

        # bulk_data = ""
        lines = []

        for item in data:
            # First row of the bulk
            if id_key is not None:
                _id = item[id_key]
                action = {method: {"_index": index_name, "_id": _id}}
            else:
                action = {method: {"_index": index_name}}

            # Second row (actual data) of the bulk
            if method == "update":
                # bulk_data += f"{json.dumps(action)}\n"
                bulk_data = json.dumps(
                    {"doc": item, "doc_as_upsert": True},
                    ensure_ascii=False,
                )
            else:
                # bulk_data = json.dumps(action)}\n{
                bulk_data = json.dumps(item, ensure_ascii=False)

            lines.append(json.dumps(action))
            lines.append(bulk_data)

        return "\n".join(lines) + "\n"

    def bulk_index(
        self,
        index_name: str,
        data: list[dict],
        id_key: str | None = None,
        routing_key: str | None = None,
        chunk_size: int = 10,
        refresh: Literal["true", "wait_for"] = "true",
    ):
        """Bulk index/update documents

        Parameters
        ----------
            - index_name (str): Index name to update to

            - data (list[dict]): A list of dictionary data to update to the index

            - id_key (str | None): Key of the dictionary data to use for documents' `_id` (Default = None)
                If None is provided, it will auto increment

            - routing_key (str | None): Routing name (shard) for all the data in this batch (Default = None)

            - chunk_size (int): The amount of data to be bulk updated in a single loop

            - refresh (Literal["true", "wait_for"]): Refresh parameter for Elasticsearch update/index API

        Returns
        -------
            Response
        """

        self._check_authen()

        es_url = ppath.join(self.es_endpoint, index_name, "_bulk")
        es_url += f"?refresh={refresh}"

        if routing_key:
            es_url += f"&routing={routing_key}"

        i = 0
        while i < len(data):
            data_chunk = data[i : i + chunk_size]
            bulk_payload = self.generate_bulk_payload(
                index_name, data_chunk, method="index", id_key=id_key
            )

            headers = self.headers.copy()
            headers["Content-Type"] = "application/json"

            # Send bulk request
            response = requests.post(es_url, headers=headers, data=bulk_payload)

            if response.status_code != 200:
                print(
                    f"Error indexing batch {i} to {i + chunk_size - 1}: {response.text}"
                )
                return response
            else:
                print(f"Successfully indexed batch {i} to {i + chunk_size - 1}")

            i += chunk_size

        print("Finished updating all batches.")
        r = RequestResponse(
            status_code=200,
            content=json.dumps({"message": "All chunks are updated successfully."}),
        )
        return r

    def bulk_update(
        self,
        index_name: str,
        data: list[dict],
        id_key: str,
        routing_key: str | None = None,
        chunk_size: int = 100,
        refresh: Literal["true", "wait_for"] = "true",
    ):
        """Bulk index/update documents

        Parameters
        ----------
            - index_name (str): Index name to update to

            - data (list[dict]): A list of dictionary data to update to the index

            - id_key (str): Key of the dictionary data to use for documents' `_id`

            - routing_key (str | None): Routing name (shard) for all the data in this batch (Default = None)

            - chunk_size (int): The amount of data to be bulk updated in a single loop

            - refresh (Literal["true", "wait_for"]): Refresh parameter for Elasticsearch update/index API

        Returns
        -------
            Response
        """

        self._check_authen()

        es_url = ppath.join(self.es_endpoint, index_name, "_bulk")
        es_url += f"?refresh={refresh}"

        if routing_key:
            es_url += f"&routing={routing_key}"

        i = 0
        while i < len(data):
            data_chunk = data[i : i + chunk_size]
            bulk_payload = self.generate_bulk_payload(
                index_name, data_chunk, method="update", id_key=id_key
            )

            headers = self.headers.copy()
            headers["Content-Type"] = "application/json"

            # Send bulk request
            response = requests.post(es_url, headers=headers, data=bulk_payload)

            if response.status_code != 200:
                print(
                    f"Error indexing batch {i} to {i + chunk_size - 1}: {response.text}"
                )
                return response
            else:
                print(f"Successfully indexed batch {i} to {i + chunk_size - 1}")

            i += chunk_size

        print("Finished updating all batches.")
        r = RequestResponse(
            status_code=200,
            content=json.dumps({"message": "All chunks are updated successfully."}),
        )
        return r

    def search(self, index_name: str, dsl: dict) -> SearchResults:
        """Search API to a speicified index and DSL query

        Parameters
        ----------
            - index_name (str): Index name to search from

            - dsl (dict): A DSL query

        Returns
        -------
            SearchResults object

        Example
        -------
            Example of using the results object retrieved from the `search()`

            ```
            dsl = {"query": {"match": ...}}
            results = els.search("my_index", dsl)

            sources = results.get_sources(include_id=True, include_score=True)
            df = results.to_dataframe(include_id=True, include_score=True)
            ```
        """

        self._check_authen()

        es_url = ppath.join(self.es_endpoint, index_name, "_search")
        headers = self.headers.copy()
        headers["Content-Type"] = "application/json"

        results = requests.post(url=es_url, headers=headers, data=json.dumps(dsl))
        search_results = SearchResults(results)
        return search_results

    def explain(
        self,
        index_name: str,
        doc_id: str,
        dsl: dict,
        routing: str | None = None,
    ) -> ExplainResult:
        """Explain API to a specified index and DSL query

        Parameters
        ----------
            - index_name (str): Index name to search from

            - doc_id (str): The document's `_id` to be explained by the DSL query

            - dsl (dict): A DSL query

            - routing (str | None): The routing (shard) name, if routing is required (Default = None)

        Returns
        -------
            ExplainResult object
        """

        self._check_authen()

        es_url = ppath.join(self.es_endpoint, index_name, "_explain", doc_id)
        params = {"routing": routing} if routing else {}

        headers = self.headers.copy()
        headers["Content-Type"] = "application/json"

        dsl = {"query": dsl["query"]}

        results = requests.get(
            es_url, params=params, headers=headers, data=json.dumps(dsl)
        )
        return ExplainResult(results)

    def count(self, index_name: str, dsl: dict | None = None):
        """Returns the documents count in the specified `index_name`

        Parameters
        ----------
        index_name: str
            Index to count from

        dsl: dict | None (Default = None)
            DSL query you want to count the results of. If None, return the whole documents count.

        Returns
        -------
        CountResults
            CountResults object
        """

        self._check_authen()

        es_url = ppath.join(self.es_endpoint, index_name, "_count")
        if not dsl or not isinstance(dsl, dict):
            results = requests.get(es_url, headers=self.headers)
            return CountResults(results)

        headers = self.headers.copy()
        headers["Content-Type"] = "application/json"

        dsl = {"query": dsl["query"]}

        results = requests.post(es_url, headers=headers, data=json.dumps(dsl))
        return CountResults(results)

    def get_cat(self) -> CatResults:
        """Returns the stats of all the indices

        The `store_size` is in mb.

        Returns
        -------
            CatResults object
        """

        self._check_authen()

        es_url = ppath.join(self.es_endpoint, "_cat/indices?format=json&bytes=mb")
        results = requests.get(es_url, headers=self.headers)

        return CatResults(results)

    def get_stats(self, index_name: str):
        """Returns the detailed stats of a specific index"""

        self._check_authen()

        es_url = ppath.join(self.es_endpoint, index_name, "_stats")
        results = requests.get(es_url, headers=self.headers)

        return StatsResults(results)

    def _check_authen(self):
        if not self.is_authen:
            raise Exception("Please authenticate first using `authen()` method.")

    def __repr__(self):
        return f"<ELS at {self.es_endpoint}, Authen: {self.is_authen}>"



import math
import re

from .types import ExplainDetailsDict, FieldScoreDict, ScoreSummaryDict


def get_scores_terms(
    dict_or_details_list: list[ExplainDetailsDict] | ExplainDetailsDict,
    results: list = None,
    field_name: str = None,
    clause: str = None,
) -> list[FieldScoreDict]:
    if results is None:
        results = []

    if isinstance(dict_or_details_list, dict):
        value = dict_or_details_list.get("value")
        description = dict_or_details_list.get("description", "")

        # Handle description like 'name.keyword:ปูน*^500.0'
        match = re.match(r"([\w\.]+):(.+?\*\^([\d.]+))", description)
        if match:
            field, clause_str, boost_str = match.groups()
            results.append(
                {
                    "field": field,
                    "clause": clause_str,
                    "type": "value",
                    "value": value,
                }
            )
            results.append(
                {
                    "field": field,
                    "clause": clause_str,
                    "type": "boost",
                    "value": float(boost_str),
                }
            )

        # Set field_name and clause only when pattern is recognized
        if "weight(" in description:
            try:
                segment = description.split("weight(")[1].split(" in ")[0]
                if ":" in segment:
                    field_name, clause = segment.split(":", 1)

            except Exception:
                pass

        elif (
            ":" in description
            and not description.lower().startswith("sum of")
            and not description.lower().startswith("max of")
            and not description.lower().startswith("avg of")
            and not field_name
        ):
            # fallback: try split on `:` if field name wasn't set
            field_name, _, clause = description.partition(":")

        # Match score types
        score_types = ("boost", "idf", "tf")
        for score_type in score_types:
            if description.lower().startswith(score_type):
                results.append(
                    {
                        "field": field_name or "unknown",
                        "clause": clause or "unknown",
                        "type": score_type,
                        "value": value,
                    }
                )

        for detail in dict_or_details_list.get("details", []):
            get_scores_terms(detail, results, field_name, clause)

    elif isinstance(dict_or_details_list, list):
        for item in dict_or_details_list:
            get_scores_terms(item, results, field_name, clause)

    return results


def get_scores_summary(
    details_list: list[FieldScoreDict],
) -> dict[str, ScoreSummaryDict]:
    ret_details = {}
    all_fields = {i["field"] for i in details_list}
    for key in all_fields:
        ret_details[key] = {}
        score_items = [i for i in details_list if i["field"] == key]

        # n_items should either contain 2, 3, or a multiple of 3 items,
        # e.g., [{'field':.., 'type':'boost', 'value':..},
        #        {'field':.., 'type':'tf', 'value':..},
        #        {'field':.., 'type':'idf', 'value':..},
        #       ]

        n_items = len(score_items)
        if n_items <= 2:  # Keyword or exact-match type, store the value and boost as is
            for si in score_items:
                ret_details[key][si["type"]] = si["value"]

        elif n_items == 3:  # boost, tf, idf
            ret_details[key]["value"] = 0
            prod = math.prod([i["value"] for i in score_items])
            ret_details[key]["value"] += prod
            _boost = list(filter(lambda x: x["type"] == "boost", score_items))[
                0
            ]  # get boost dict
            ret_details[key]["boost"] = _boost["value"]

        elif n_items > 3:  # multiple boost, tf, idf for different clauses
            field_clauses = {(i["field"], i["clause"]) for i in score_items}
            ret_details[key]["value"] = 0
            max_boost = max([i["value"] for i in score_items if i["type"] == "boost"])

            for fc_i in field_clauses:
                scores_to_be_multiplied = [
                    i["value"]
                    for i in score_items
                    if i["field"] == fc_i[0] and i["clause"] == fc_i[1]
                ]

                # accumulate for the sub instances, for the same field name, it can have many matches with multiple boost, tf, idf
                # we will capture in a multiple of 3
                n_chunks = len(scores_to_be_multiplied) // 3
                chunk_i = 0
                _sum = 0
                while (
                    chunk_i < n_chunks
                ):  # chunk_i runs from 0, 1, 2 in case len()==8 (3 chunks)
                    _sum += math.prod(
                        scores_to_be_multiplied[chunk_i * 3 : (chunk_i + 1) * 3]
                    )
                    chunk_i += 1

                ret_details[key]["value"] += _sum
                ret_details[key]["boost"] = max_boost

    return ret_details


# Test new extraction functions


def get_field_and_clause(text_details):
    if match := re.search(r"weight\((\w+):(\w+) in", text_details):
        field = match[1]
        clause = match[2]

        return field, clause


def get_contribution_details(break_down):
    cur_depth = 0
    results = []
    for i, (depth, score, details) in enumerate(break_down):
        if details in ("sum of:", "avg of:", "max of:", "min of:"):
            cur_depth = depth  # Set the current depth level
            continue  # Move on to its details

        if depth == cur_depth + 1:  # one level deeper than the current aggregate
            if details.startswith("weight("):
                try:
                    field, clause = get_field_and_clause(details)

                except Exception:
                    print(details)

                for j in range(i, len(break_down)):
                    dtl_depth = break_down[j][0]
                    dtl_score = break_down[j][1]
                    dtl_detail = break_down[j][2]

                    if (
                        dtl_depth == depth + 1 and dtl_detail == "boost"
                    ):  ## one level deeper than the main score
                        boost = dtl_score
                        break

                results.append([field, clause, score, boost])

    return results



from typing import Literal, TypedDict, NotRequired


class ExplainDetailsDict(TypedDict):
    value: float
    description: str
    details: list["ExplainDetailsDict"]


class FieldScoreDict(TypedDict):
    field: str
    clause: str
    type: Literal[r"value", "boost", "idf", "tf"]
    value: int | float


class ScoreSummaryDict(TypedDict):
    value: float
    boost: float


# Results from _cat
class CatDict(TypedDict):
    health: str
    status: str
    index: str
    uuid: str
    pri: str
    rep: str
    docs_count: str
    docs_deleted: str
    store_size: str
    pri_store_size: str


# Results from _stats
class DocsStats(TypedDict):
    count: int
    deleted: int


class StoreStats(TypedDict):
    size_in_bytes: int
    total_data_set_size_in_bytes: int
    reserved_in_bytes: int


class IndexingStats(TypedDict):
    index_total: int
    index_time_in_millis: int
    index_current: int
    index_failed: int
    delete_total: int
    delete_time_in_millis: int
    delete_current: int
    noop_update_total: int
    is_throttled: bool
    throttle_time_in_millis: int


class GetStats(TypedDict):
    total: int
    time_in_millis: int
    exists_total: int
    exists_time_in_millis: int
    missing_total: int
    missing_time_in_millis: int
    current: int


class SearchStats(TypedDict):
    open_contexts: int
    query_total: int
    query_time_in_millis: int
    query_current: int
    fetch_total: int
    fetch_time_in_millis: int
    fetch_current: int
    scroll_total: int
    scroll_time_in_millis: int
    scroll_current: int
    suggest_total: int
    suggest_time_in_millis: int
    suggest_current: int


class MergesStats(TypedDict):
    current: int
    current_docs: int
    current_size_in_bytes: int
    total: int
    total_time_in_millis: int
    total_docs: int
    total_size_in_bytes: int
    total_stopped_time_in_millis: int
    total_throttled_time_in_millis: int
    total_auto_throttle_in_bytes: int


class RefreshStats(TypedDict):
    total: int
    total_time_in_millis: int
    external_total: int
    external_total_time_in_millis: int
    listeners: int


class FlushStats(TypedDict):
    total: int
    periodic: int
    total_time_in_millis: int


class WarmerStats(TypedDict):
    current: int
    total: int
    total_time_in_millis: int


class QueryCacheStats(TypedDict):
    memory_size_in_bytes: int
    total_count: int
    hit_count: int
    miss_count: int
    cache_size: int
    cache_count: int
    evictions: int


class FieldDataStats(TypedDict):
    memory_size_in_bytes: int
    evictions: int


class SegmentsStats(TypedDict):
    count: int
    memory_in_bytes: int
    terms_memory_in_bytes: int
    stored_fields_memory_in_bytes: int
    term_vectors_memory_in_bytes: int
    norms_memory_in_bytes: int
    points_memory_in_bytes: int
    doc_values_memory_in_bytes: int
    index_writer_memory_in_bytes: int
    version_map_memory_in_bytes: int
    fixed_bit_set_memory_in_bytes: int
    max_unsafe_auto_id_timestamp: int
    file_sizes: dict[str, int]


class TranslogStats(TypedDict):
    operations: int
    size_in_bytes: int
    uncommitted_operations: int
    uncommitted_size_in_bytes: int
    earliest_last_modified_age: int


class RequestCacheStats(TypedDict):
    memory_size_in_bytes: int
    evictions: int
    hit_count: int
    miss_count: int


class RecoveryStats(TypedDict):
    current_as_source: int
    current_as_target: int
    throttle_time_in_millis: int


class ShardsMetadata(TypedDict):
    total: int
    successful: int
    failed: int


class ShardStats(TypedDict):
    total_count: int


class StatsBlock(TypedDict):
    docs: DocsStats
    shard_stats: ShardStats
    store: StoreStats
    indexing: IndexingStats
    get: GetStats
    search: SearchStats
    merges: MergesStats
    refresh: RefreshStats
    flush: FlushStats
    warmer: WarmerStats
    query_cache: QueryCacheStats
    fielddata: FieldDataStats
    completion: dict[str, int]
    segments: SegmentsStats
    translog: TranslogStats
    request_cache: RequestCacheStats
    recovery: RecoveryStats


class IndexStats(TypedDict):
    uuid: NotRequired[str]
    primaries: StatsBlock
    total: StatsBlock


class ShardsStats(TypedDict):
    total: int
    successful: int
    failed: int


class StatsResultsDict(TypedDict):
    _shards: ShardsMetadata
    _all: IndexStats
    indices: dict[str, IndexStats]


# count results
class CountShardsDict(TypedDict):
    total: int
    successful: int
    skipped: int
    failed: int


class CountDict(TypedDict):
    count: int
    _shards: CountShardsDict
